#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <conio.h>

/*
#define _DEBUG_
*/

#define IN_SIZE 1024

#define BUF_SIZE 1024
#define BUF_MASK ((unsigned long)0xFFFFFC00)
#define BUF_INVT ((unsigned long)0x000003FF)

typedef struct _t_Target {
	int handle;
	char buffer[BUF_SIZE];
	int count, modify;
	unsigned long size, seek;
} TARGET;

int process_file(char *pstr, int line, int len, TARGET *t)
{
	char nb[IN_SIZE];

	if(t->handle != -1) {
		if(t->count && t->modify) {
			lseek(t->handle, t->seek, SEEK_SET);
			write(t->handle, t->buffer, t->count);
		}
		close(t->handle);
		t->handle = -1;
		printf("\n");
	}

	if(sscanf(pstr+8, "%s", nb) == 1) {
		t->handle = open(nb, O_RDWR | O_BINARY);
		if(t->handle != -1) {
			t->size = lseek(t->handle, 0, SEEK_END);
			t->seek = (t->size + BUF_SIZE -1) & BUF_MASK;
			t->count = 0;
			t->modify = 0;
			printf("FILENAME %s\n", nb);
		}
		else {
			printf("%s\n", pstr);
			printf("--> line(%d) error \"can't open file\"\n\n", line);
		}
	}

	return t->handle == -1 ? 1 : 0;
}

int process_patch(char *pstr, int line, int len, TARGET *t)
{
	unsigned long cur, ofs;
	unsigned int org, mod;

	if(sscanf(pstr, "%lx:%x%x", &cur, &org, &mod) == 3) {
		if(t->handle != -1) {
			if(cur < t->size) {
				ofs = cur & BUF_INVT;
				cur &= BUF_MASK;

				if(t->seek != cur) {
					if(t->count && t->modify) {
						lseek(t->handle, t->seek, SEEK_SET);
						write(t->handle, t->buffer, t->count);
					}
					t->seek = lseek(t->handle, cur, SEEK_SET);
					t->count = read(t->handle, t->buffer, BUF_SIZE);
					t->modify = 0;
					#ifdef _DEBUG_
					printf("pointer seek to %08lX (%08lX) / count %d\n", cur, t->seek, t->count);
					#endif
				}

				#ifdef _DEBUG_
				printf("cur : %08lX   ofs : %08lX  ", cur, ofs);
				#endif
				if(ofs < (unsigned int)t->count && (unsigned int)t->buffer[ofs] == org) {
					#ifdef _DEBUG_
					printf("--> res1:%08lX: %02X %02X ofs(%08lX) count(%08X)\n", cur + ofs, org, mod, ofs, t->count);
					#endif
					#ifndef _DEBUG_
					t->buffer[ofs] = mod;
					t->modify = 1;
					#endif
				}
				else {
					#ifdef _DEBUG_
					printf("--> res2:%08lX: %02X %02X ofs(%08lX) count(%08lX)\n", cur + ofs, org, mod, ofs, t->count);
					#endif
					#ifndef _DEBUG_
					printf("%s\n", pstr);
					printf("--> line(%d) error \"don't match original data\"\n", line);
					#endif
				}
			}
			else {
				printf("%s\n", pstr);
				printf("--> line(%d) error \"address not found\"\n", line);
			}
		}
		else {
			printf("%s\n", pstr);
			printf("--> line(%d) error \"target file not specified\"\n", line);
		}
		return 1;
	}
	return 0;
}

void process_line(char *pstr, int line, int len, TARGET *t)
{
	static int comment = 0;

	if(*pstr == '*') {
		printf("%s\n", pstr);
		comment = 1;
		return;
	}
	else if(comment) {
		printf("\n");
		comment = 0;
	}

	if(len >= 8 && !strnicmp(pstr, "filename", 8) && process_file(pstr, line, len, t)) return;
	if(t->handle != -1 && process_patch(pstr, line, len, t)) return;
}

int process_main(FILE *pfile)
{
	int len, line;
	char in_str[IN_SIZE], *in_ptr;
	TARGET *t;

	t = (TARGET *)malloc(sizeof(TARGET));
	if(!t) {
		printf("\nMemory Error\n");
		exit(0);
	}
	t->handle = -1;
	t->count = t->modify = 0;
	t->size = t->seek = 0;

	for(line=0; fgets(in_str, IN_SIZE -1, pfile) == in_str; line++) {
		len = strlen(in_str);
		in_ptr = in_str + (len -1);
		while(in_ptr >= in_str && (*in_ptr == '\n' || *in_ptr == ' ' || *in_ptr == '\t')) in_ptr--, len--;
		*(in_ptr + 1) = '\0';
		in_ptr = in_str;
		while(*in_ptr == ' ' || *in_ptr == '\t') in_ptr++, len--;

		process_line(in_ptr, line, len, t);
	}

	if(t->handle != -1) {
		if(t->count && t->modify) {
			lseek(t->handle, t->seek, SEEK_SET);
			write(t->handle, t->buffer, t->count);
		}
		close(t->handle);
		t->handle = -1;
		printf("\n");
	}

	if(ferror(pfile)) return 0;

	free((void *)t);
	return 1;
}

void main(int argc, char **argv)
{
	FILE *pfile;
	int result;

	printf("Patch Utility  00/06/10(yy/mm/dd)\n");
	printf("\n");

	if(argc < 2) {
		printf("Usage: patch <patch_file_name>\n");
		exit(0);
	}

	pfile = fopen(argv[1], "r");
	if(!pfile) {
		printf("Usage: patch <patch_file_name>\n");
		exit(0);
	}
	result = process_main(pfile);
	fclose(pfile);

	if(result) {
		printf("\n");
	}
	else {
		printf("%s File Read Error\n", argv[1]);
		exit(0);
	}
}
